<?php

namespace Drupal\basicshib;

use Drupal\basicshib\Exception\AttributeException;
use Drupal\basicshib\Exception\AuthenticationException;
use Drupal\basicshib\Plugin\AuthFilterPluginInterface;
use Drupal\basicshib\Plugin\BasicShibPluginManager;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Path\CurrentPathStack;
use Drupal\Core\Path\PathValidatorInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\mymodule\PathProcessor\BasicShibPathProcessor;
 use Drupal\user\UserInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\Url;
use Drupal\Core\Routing\RedirectDestination;

/**
 *
 */
class AuthenticationHandler implements AuthenticationHandlerInterface {
  /**
   * Instance variable attribute mapper.
   *
   * @var AttributeMapperInterface
   */
  private AttributeMapperInterface $attribute_mapper;
  /**
   * Instance variable session tracker.
   *
   * @var SessionTracker
   */
  private SessionTracker $session_tracker;
  /**
   * Instance variable user provider.
   *
   * @var \Drupal\basicshib\Plugin\UserProviderPluginInterface
   */
  private $user_provider;
  /**
   * Instance variable auth filter .
   *
   * @var \Drupal\basicshib\Plugin\AuthFilterPluginInterface[]
   */
  private array $auth_filters = [];

  /**
   * Instance variable handlers.
   *
   * @var array
   */
  private $handlers;

  /**
   * Instance variable attributes.
   *
   * @var array
   */
  private $attributes;

  /**
   * Instance variable path validator.
   *
   * @var \Drupal\Core\Path\PathValidatorInterface
   */
  private PathValidatorInterface $path_validator;
  /**
   * Instance variable redirect destination.
   *
   * @var \Drupal\Core\Routing\RedirectDestination
   */
  private RedirectDestination $redirect_destination;

  private $default_login_redirect_destination;
  /**
   * Instance variable path validator.
   *
   * @var \Drupal\Core\Path\CurrentPathStack
   */
  private CurrentPathStack $current_path;

  /**
   * @var \Psr\Log\LoggerInterface
   */

  /**
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   * @param \Drupal\basicshib\AttributeMapperInterface $attribute_mapper
   * @param \Drupal\basicshib\Plugin\BasicShibPluginManager $user_provider_plugin_manager
   * @param \Drupal\basicshib\Plugin\BasicShibPluginManager $auth_filter_plugin_manager
   * @param \Drupal\Core\Path\PathValidatorInterface $path_validator
   * @param \Drupal\Core\Routing\RedirectDestination $redirect_destination
   * @param \Drupal\Core\Path\CurrentPathStack $current_path
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    RequestStack $request_stack,
//    AttributeMapperInterface $attribute_mapper,
    BasicShibPluginManager $user_provider_plugin_manager,
//    BasicShibPluginManager $auth_filter_plugin_manager,
    PathValidatorInterface $path_validator,
    RedirectDestination $redirect_destination,
    CurrentPathStack $current_path
  ) {
    $authentication = $config_factory->get('basicshib.authentication.default_settings');
    $this->default_login_redirect_destination = $authentication->get('default_post_login_redirect_path');
    $this->handlers = [
      'login' => $authentication->get('login_handler'),
      'logout' => $authentication->get('logout_handler'),
    ];

    $this->attributes = [
      'name' => $authentication->get('name_attribute'),
      'mail' => $authentication->get('mail_attribute'),
      'session_id' => $authentication->get('session_id_attribute'),
    ];

    $auth_filter = $config_factory->get('basicshib.auth_filter');

    if ($request_stack->getCurrentRequest()->hasSession()) {
      $this->session_tracker = new SessionTracker(
        $request_stack->getCurrentRequest()->getSession()
      );
    }

    $plugins = $config_factory->get('basicshib.settings')->get('plugins');
//    $this->attribute_mapper = $attribute_mapper;
    $this->user_provider = $user_provider_plugin_manager
      ->createInstance($plugins['user_provider']);
//    foreach ($plugins['auth_filter'] as $name) {
//      $this->auth_filters[$name]
//        = $auth_filter_plugin_manager->createInstance($name);
//    }


    $this->path_validator = $path_validator;
    $this->redirect_destination = $redirect_destination;
    $this->current_path = $current_path;
  }

  /**
   * {@inheritDoc}
   * @throws \Drupal\basicshib\Exception\AttributeException
   */
  public function authenticate() {
    // Get key attributes.
    $key_attributes = $this->getKeyAttributes();
    // Load the user account.
    $account = $this->user_provider->loadUserByName($key_attributes['name']);
    if ($account) {
      $this->assertExistingUserLoginAllowed($account);
    }
    else {
      $this->assertUserCreationAllowed($key_attributes['name']);
      $account = $this->user_provider
        ->createUser($key_attributes['name'], $key_attributes['mail']);
    }

    $this->saveAccount($account, $key_attributes);
    $this->session_tracker->set($key_attributes['session_id']);
    $this->userLoginFinalize($account);

    return $account;
  }

  /**
   * Save account function.
   *
   * @param \Drupal\user\UserInterface $account
   * @param array $key_attributes
   *
   * @throws AuthenticationException
   */
  private function saveAccount(UserInterface $account, array $key_attributes) {
    $message = $account->isNew() ? 'Saving new account for \'%s\' has failed' : 'Updating existing account for \'%s\' has failed';
    $code = $account->isNew() ? AuthenticationException::USER_CREATION_FAILED : AuthenticationException::USER_UPDATE_FAILED;
    if ($account->isNew()) {
      try {
        $account->save();
      }
      catch (EntityStorageException $exception) {
        throw new AuthenticationException(sprintf($message, $account->getAccountName()), $code, $exception);
      }
    }
  }

  /**
   * Assert that an existing user is allowed to log in.
   *
   * @param \Drupal\user\UserInterface $account
   *
   * @throws AuthenticationException
   */
  private function assertExistingUserLoginAllowed(UserInterface $account) {
    if ($account->isBlocked()) {
      throw new AuthenticationException(sprintf('User \'%s\' is blocked and cannot be authenticated', $account->getAccountName()), AuthenticationException::USER_BLOCKED);
    }

    foreach ($this->auth_filters as $auth_filter) {
      if (!$auth_filter->isExistingUserLoginAllowed($account)) {
        throw new AuthenticationException($auth_filter->getError(AuthFilterPluginInterface::ERROR_EXISTING_NOT_ALLOWED, $account), AuthenticationException::LOGIN_DISALLOWED_FOR_USER);
      }
    }
  }

  /**
   * Assert that user creation is allowed.
   *
   * @throws AuthenticationException
   */
  private function assertUserCreationAllowed($name) {
    foreach ($this->auth_filters as $auth_filter) {
      if (!$auth_filter->isUserCreationAllowed()) {
        $message = $auth_filter->getError(AuthFilterPluginInterface::ERROR_CREATION_NOT_ALLOWED);
        throw new AuthenticationException(sprintf('%s, user=\'%s\'', $message, $name), AuthenticationException::USER_CREATION_NOT_ALLOWED);
      }
    }
  }

  /**
   * Get the key attributes 'session_id', 'name', and 'mail'
   *
   * @return array
   *   An associative array whose keys are:
   *   - session_id
   *   - name
   *   - mail
   *
   * @throws \Drupal\basicshib\Exception\AttributeException
   */
  private function getKeyAttributes(): array {
    // #### LOCAL TEST CODE
    // EXISTING USER
    $attributes = [
      'session_id' => 'qwerty',
      'name' => 'admin',
      'mail' => 'mgd20@duke.edu',
    ];
    // Blocked User
//    $attributes = [
//      'session_id' => '1234',
//      'name' => 'me',
//      'mail' => 'me@duke.edu',
//    ];
    // NEW USER
//    $attributes = [
//      'session_id' => 'abcd',
//      'name' => 'jdoe@example.com',
//      'mail' => 'jdoe@mail.example.com',
//    ];
    // ### LIVE CODE
    return $attributes;
  }

  /**
   * Checks user session and logs the user out if the session is not valid.
   *
   * @param Request $request
   * @param AccountProxyInterface $account
   *
   * @return int
   */
  public function checkUserSession(Request $request, AccountProxyInterface $account): int {
    // Handle anonymous user.
    if ($account->isAnonymous()) {
      if ($this->session_tracker->exists()) {
        $this->session_tracker->clear();
        return self::AUTHCHECK_LOCAL_SESSION_EXPIRED;
      }
      return self::AUTHCHECK_IGNORE;
    }

    // Authenticated user who is not authenticated via shibboleth.
    if (!$this->session_tracker->exists()) {
      return self::AUTHCHECK_IGNORE;
    }

    // Authenticated user with expired shib session.
    $session_id = $this->attributes['session_id'];
    if (!$session_id) {
      $this->terminateSession($account);
      return self::AUTHCHECK_SHIB_SESSION_EXPIRED;
    }

    // Authenticated user whose tracked session id does not match the current
    // Session id.
    if ($session_id !== $this->session_tracker->get()) {
      $this->terminateSession($account);
      return self::AUTHCHECK_SHIB_SESSION_ID_MISMATCH;
    }

    // Additional checks by auth filter plugins.
    foreach ($this->auth_filters as $auth_filter) {
      $value = $auth_filter->checkSession($request, $account);
      if ($value !== self::AUTHCHECK_IGNORE) {
        $this->terminateSession($account);
        return $value;
      }
    }

    $this->setLoginUrl();
    return self::AUTHCHECK_IGNORE;
  }

  /**
   * Terminates the current session
   *
   * @param AccountProxyInterface $account
   */
  private function terminateSession(AccountProxyInterface $account) {
    $this->session_tracker->clear();
    user_logout();
  }

  /**
   * Finalize login.
   *
   * @param \Drupal\user\UserInterface $account
   */
  private function userLoginFinalize(UserInterface $account) {
    user_login_finalize($account);
  }

  /**
   * {@inheritDoc}
   *
   * @return String login URL
   */
  public function setLoginUrl(): Url {
    $loginHandler = $this->handlers['login'];
    $loginUrl = Url::fromUserInput($loginHandler);

    $targetUrl = $this->path_validator->getUrlIfValid('/basicshib/login');
    $targetUrl->setAbsolute();

    $afterLogin = $this->default_login_redirect_destination;
    if ($afterLogin == 'current') {
      $referer = \Drupal::request()->server->get('HTTP_REFERER');
      $baseUrl = \Drupal::request()->createFromGlobals()->getSchemeAndHttpHost();
      $afterLogin = substr($referer, strlen($baseUrl));
    }
    $targetUrl->setOption('query', ['after_login' => $afterLogin]);
    $loginUrl->setOption('query', ['target' => $targetUrl->toString()]);
    return $loginUrl;

  }

}
